@@ -3,7 +3,7 @@
 #
 # furl - URL manipulation made simple.
 #
-# Arthur Grunseid
+# Ansgar Grunseid
 # grunseid.com
 # grunseid@gmail.com
 #
@@ -21,8 +21,8 @@ from six.moves.urllib.parse import quote, unquote, quote_plus
 
 from .omdict1D import omdict1D
 from .compat import basestring, UnicodeMixin
-
-_absent = object()
+from .common import (
+    callable_attr, is_iterable_but_not_string, absent as _absent)
 
 #
 # TODO(grun): Subclass Path, PathCompositionInterface, Query, and
@@ -51,18 +51,6 @@ DEFAULT_PORTS = {
     'telnet': 23,
 }
 
-# List of schemes that don't require two slashes after the colon. For example,
-# 'mailto:user@google.com' instead of 'mailto://user@google.com'. Scheme
-# strings are lowercase.
-#
-# TODO(grun): Support schemes separated by just ':', not '://' without having
-# an explicit list. There are many such schemes in various URIs.
-COLON_SEPARATED_SCHEMES = [
-    'sms',
-    'tel',
-    'mailto',
-]
-
 
 def lget(l, index, default=None):
     try:
@@ -161,7 +149,7 @@ def static_vars(**kwargs):
 #   scheme      = ALPHA *( ALPHA / DIGIT / "+" / "-" / "." )
 #
 PERCENT_REGEX = r'\%[a-fA-F\d][a-fA-F\d]'
-INVALID_DOMAIN_CHARS = '!@#$%^&\'\"*()+=:;/'
+INVALID_HOST_CHARS = '!@#$%^&\'\"*()+=:;/'
 
 
 @static_vars(regex=re.compile(
@@ -187,76 +175,93 @@ def is_valid_scheme(scheme):
     return is_valid_scheme.regex.match(scheme) is not None
 
 
-@static_vars(regex=re.compile('[%s]' % re.escape(INVALID_DOMAIN_CHARS)))
-def is_valid_domain(domain):
-    toks = domain.split('.')
+@static_vars(regex=re.compile('[%s]' % re.escape(INVALID_HOST_CHARS)))
+def is_valid_host(hostname):
+    toks = hostname.split('.')
     if toks[-1] == '':  # Trailing '.' in a fully qualified domain name.
         toks.pop()
 
     for tok in toks:
-        if is_valid_domain.regex.search(tok) is not None:
+        if is_valid_host.regex.search(tok) is not None:
             return False
 
     return '' not in toks  # Adjacent periods aren't allowed.
 
 
-def _get_scheme(url):
-    if url.lstrip().startswith('//'):  # Protocol relative URL.
+def get_scheme(url):
+    if url.startswith(':'):
         return ''
 
-    before_colon = url[:max(0, url.find(':'))]
-    if before_colon in COLON_SEPARATED_SCHEMES:
-        scheme = before_colon
-    else:
-        scheme = url[:max(0, url.find('://'))] or None
-    return scheme if (scheme is not None and is_valid_scheme(scheme)) else None
+    # Avoid incorrect scheme extraction with url.find(':') when other URL
+    # components, like the path, query, fragment, etc, may have a colon in
+    # them. For example, the URL 'a?query:', whose query has a ':' in it.
+    no_fragment = url.split('#', 1)[0]
+    no_query = no_fragment.split('?', 1)[0]
+    no_path_or_netloc = no_query.split('/', 1)[0]
+    scheme = url[:max(0, no_path_or_netloc.find(':'))] or None
+
+    if scheme is not None and not is_valid_scheme(scheme):
+        return None
+
+    return scheme
 
 
-def _set_scheme(url, newscheme):
-    scheme = _get_scheme(url)
-    newscheme = newscheme or ''
-    newseparator = ':' if newscheme in COLON_SEPARATED_SCHEMES else '://'
-    if scheme == '':  # Protocol relative URL.
-        url = '%s:%s' % (newscheme, url)
-    elif scheme is None and url:  # No scheme.
-        url = ''.join([newscheme, newseparator, url])
-    elif scheme:  # Existing scheme.
-        remainder = url[len(scheme):]
-        if remainder.startswith('://'):
-            remainder = remainder[3:]
-        elif remainder.startswith(':'):
-            remainder = remainder[1:]
-        url = ''.join([newscheme, newseparator, remainder])
+def strip_scheme(url):
+    scheme = get_scheme(url) or ''
+    url = url[len(scheme):]
+    if url.startswith(':'):
+        url = url[1:]
     return url
 
 
+def set_scheme(url, scheme):
+    after_scheme = strip_scheme(url)
+    if scheme is None:
+        return after_scheme
+    else:
+        return '%s:%s' % (scheme, after_scheme)
+
+
 def urlsplit(url):
     """
     Parameters:
       url: URL string to split.
-
     Returns: urlparse.SplitResult tuple subclass, just like
-    urlparse.urlsplit() returns, with fields (scheme, netloc, path,
-    query, fragment, username, password, hostname, port). See the url
-    below for more details on urlsplit().
-
-      http://docs.python.org/library/urlparse.html#urlparse.urlsplit
+      urlparse.urlsplit() returns, with fields (scheme, netloc, path,
+      query, fragment, username, password, hostname, port). See
+        http://docs.python.org/library/urlparse.html#urlparse.urlsplit
+      for more details on urlsplit().
     """
-    original_scheme = _get_scheme(url)
+    original_scheme = get_scheme(url)
 
-    def _change_urltoks_scheme(tup, scheme):
-        toks = list(tup)
-        toks[0] = scheme
-        return tuple(toks)
-
-    # urlsplit() only parses the query for schemes in urlparse.uses_query,
-    # so switch to 'http', a scheme in urlparse.uses_query, for
-    # urlparse.urlsplit() then restore the original scheme afterwards.
+    # urlsplit() parses URLs differently depending on whether or not the URL's
+    # scheme is in any of
+    #
+    #   urllib.parse.uses_fragment
+    #   urllib.parse.uses_netloc
+    #   urllib.parse.uses_params
+    #   urllib.parse.uses_query
+    #   urllib.parse.uses_relative
+    #
+    # For consistent URL parsing, switch the URL's scheme to 'http', a scheme
+    # in all of the aforementioned uses_* lists, and afterwards revert to the
+    # original scheme (which may or may not be in some, or all, of the the
+    # uses_* lists).
     if original_scheme is not None:
-        url = _set_scheme(url, 'http')
-    toks = urllib.parse.urlsplit(url)
-    toks_orig_scheme = _change_urltoks_scheme(toks, original_scheme)
-    return urllib.parse.SplitResult(*toks_orig_scheme)
+        url = set_scheme(url, 'http')
+
+    scheme, netloc, path, query, fragment = urllib.parse.urlsplit(url)
+
+    # Detect and preserve the '//' before the netloc, if present. E.g. preserve
+    # URLs like 'http:', 'http://', and '///sup' correctly.
+    after_scheme = strip_scheme(url)
+    if after_scheme.startswith('//'):
+        netloc = netloc or ''
+    else:
+        netloc = None
+
+    scheme = original_scheme
+    return urllib.parse.SplitResult(scheme, netloc, path, query, fragment)
 
 
 def urljoin(base, url):
@@ -267,11 +272,17 @@ def urljoin(base, url):
 
     Returns: The resultant URL from joining <base> and <url>.
     """
-    base_scheme, url_scheme = urlsplit(base).scheme, urlsplit(url).scheme
-    httpbase = _set_scheme(base, 'http')
-    joined = urllib.parse.urljoin(httpbase, url)
+    if not base:
+        return url
+
+    base_scheme = get_scheme(base)
+    url_scheme = get_scheme(url)
+
+    http_base = set_scheme(base, 'http')
+    joined = urllib.parse.urljoin(http_base, url)
     if not url_scheme:
-        joined = _set_scheme(joined, base_scheme)
+        joined = set_scheme(joined, base_scheme)
+
     return joined
 
 
@@ -317,9 +328,9 @@ def remove_path_segments(segments, remove):
     segments <segments>.
 
     Examples:
-      # '/a/b/c' - 'b/c' == '/a/'
+      # ('/a/b/c', 'b/c') -> '/a/'
       remove_path_segments(['','a','b','c'], ['b','c']) == ['','a','']
-      # '/a/b/c' - '/b/c' == '/a'
+      # ('/a/b/c', '/b/c') -> '/a'
       remove_path_segments(['','a','b','c'], ['','b','c']) == ['','a']
 
     Returns: The list of all remaining path segments after the segments
@@ -407,10 +418,10 @@ class Path(object):
         """
         if not path:
             segments = []
-        elif callable_attr(path, 'split'):  # String interface.
-            segments = self._segments_from_path(path)
-        else:  # List interface.
+        elif is_iterable_but_not_string(path):  # List interface.
             segments = path
+        else:  # String interface.
+            segments = self._segments_from_path(path)
 
         if self._force_absolute(self):
             self._isabsolute = True if segments else False
@@ -431,8 +442,9 @@ class Path(object):
 
         Returns: <self>.
         """
-        newsegments = path  # List interface.
-        if callable_attr(path, 'split'):  # String interface.
+        if is_iterable_but_not_string(path):  # List interface.
+            newsegments = path
+        else:  # String interface.
             newsegments = self._segments_from_path(path)
 
         # Preserve the opening '/' if one exists already (self.segments
@@ -455,8 +467,9 @@ class Path(object):
         if path is True:
             self.load('')
         else:
-            segments = path  # List interface.
-            if isinstance(path, six.string_types):  # String interface.
+            if is_iterable_but_not_string(path):  # List interface.
+                segments = path
+            else:  # String interface.
                 segments = self._segments_from_path(path)
             base = ([''] if self.isabsolute else []) + self.segments
             self.load(remove_path_segments(base, segments))
@@ -549,6 +562,10 @@ class Path(object):
 
         Raises: UserWarning if <path> is an improperly encoded path
         string and self.strict is True.
+
+        TODO(grun): Accept both list values and string values and
+        refactor the list vs string interface testing to this common
+        method.
         """
         segments = []
         for segment in path.split('/'):
@@ -651,8 +668,8 @@ class URLPathCompositionInterface(PathCompositionInterface):
 
     A URL path's isabsolute attribute is mutable if there's no
     netloc. The scheme doesn't matter. For example, the isabsolute
-    attribute of the URL path in 'mailto:user@domain.com', with scheme
-    'mailto' and path 'user@domain.com', is mutable because there is no
+    attribute of the URL path in 'mailto:user@host.com', with scheme
+    'mailto' and path 'user@host.com', is mutable because there is no
     netloc. See
 
       http://en.wikipedia.org/wiki/URI_scheme#Examples
@@ -710,7 +727,7 @@ class Query(object):
     The one dimensional aspect of omdict1D means that a list of values
     is interpreted as multiple values, not a single value which is
     itself a list of values. This is a reasonable distinction to make
-    because URL query parameters are one dimensional - query parameter
+    because URL query parameters are one dimensional: query parameter
     values cannot themselves be composed of sub-values.
 
     So what does this mean? This means we can safely interpret
@@ -735,17 +752,51 @@ class Query(object):
 
       str(f.query) == 'arg=one&arg=two&arg=three'
 
+
+    Additionally, while the set of allowed characters in URL queries is
+    defined in RFC 3986 section 3.4, the format for encoding key=value
+    pairs within the query is not. In turn, the parsing of encoded
+    key=value query pairs differs between implementations.
+
+    As a compromise to support equal signs in both key=value pair
+    encoded queries, like
+
+      https://www.google.com?a=1&b=2
+
+    and non-key=value pair encoded queries, like
+
+      https://www.google.com?===3===
+
+    equal signs are percent encoded in key=value pairs where the key is
+    non-empty, e.g.
+
+      https://www.google.com?equal-sign=%3D
+
+    but not encoded in key=value pairs where the key is empty, e.g.
+
+      https://www.google.com?===equal=sign===
+
+    This presents a reasonable compromise to accurately reproduce
+    non-key=value queries with equal signs while also still percent
+    encoding equal signs in key=value pair encoded queries, as
+    expected. See
+
+      https://github.com/gruns/furl/issues/99
+
+    for more details.
+
     Attributes:
       params: Ordered multivalue dictionary of query parameter key:value
-        pairs. Parameters in self.params are maintained URL decoded - 'a
-        b' not 'a+b'.
+        pairs. Parameters in self.params are maintained URL decoded,
+        e.g. 'a b' not 'a+b'.
       strict: Boolean whether or not UserWarnings should be raised if
         improperly encoded query strings are provided to methods that
         take such strings, like load(), add(), set(), remove(), etc.
     """
 
     SAFE_KEY_CHARS = "/?:@-._~!$'()*,"
-    SAFE_VALUE_CHARS = "/?:@-._~!$'()*,="
+    SAFE_VALUE_CHARS_NON_EMPTY_KEY = SAFE_KEY_CHARS
+    SAFE_VALUE_CHARS_EMPTY_KEY = SAFE_KEY_CHARS + '='
 
     def __init__(self, query='', strict=False):
         self.strict = strict
@@ -755,7 +806,8 @@ class Query(object):
         self.load(query)
 
     def load(self, query):
-        self.params.load(self._items(query))
+        items = self._items(query)
+        self.params.load(items)
         return self
 
     def add(self, args):
@@ -844,13 +896,24 @@ class Query(object):
         for key, value in self.params.iterallitems():
             utf8key = utf8(key, utf8(attemptstr(key)))
             utf8value = utf8(value, utf8(attemptstr(value)))
+
             quoted_key = quote_func(utf8key, self.SAFE_KEY_CHARS)
-            quoted_value = quote_func(utf8value, self.SAFE_VALUE_CHARS)
-            pair = '='.join([quoted_key, quoted_value])
-            if value is None:  # Example: http://sprop.su/?param
+            if not quoted_key:
+                safe_value_chars = self.SAFE_VALUE_CHARS_EMPTY_KEY
+            else:
+                safe_value_chars = self.SAFE_VALUE_CHARS_NON_EMPTY_KEY
+            quoted_value = quote_func(utf8value, safe_value_chars)
+
+            if value is None:  # Example: http://sprop.su/?param.
                 pair = quoted_key
+            else:
+                pair = '='.join([quoted_key, quoted_value])
+
             pairs.append(pair)
-        return delimiter.join(pairs)
+
+        query = delimiter.join(pairs)
+
+        return query
 
     def asdict(self):
         return {
@@ -908,10 +971,10 @@ class Query(object):
         elif callable_attr(items, 'iterallitems'):
             items = list(items.iterallitems())
         # Dictionary-like interface. e.g. {'a':1, 'b':2, 'c':3}
-        elif callable_attr(items, 'items'):
-            items = list(six.iteritems(items))
         elif callable_attr(items, 'items'):
             items = list(items.items())
+        elif callable_attr(items, 'iteritems'):
+            items = list(items.iteritems())
         # Encoded query string. e.g. 'a=1&b=2&c=3'
         elif isinstance(items, six.string_types):
             items = self._extract_items_from_querystr(items)
@@ -923,27 +986,31 @@ class Query(object):
         return items
 
     def _extract_items_from_querystr(self, querystr):
-        pairstrs = [s2 for s1 in querystr.split('&') for s2 in s1.split(';')]
-
-        if self.strict:
-            pairs = [item.split('=', 1) for item in pairstrs]
-            pairs = [(p[0], '' if len(p) == 1 else p[1]) for p in pairs]
-            for key, value in pairs:
-                valid_key = is_valid_encoded_query_key(key)
-                valid_value = is_valid_encoded_query_value(value)
-                if not valid_key or not valid_value:
-                    s = ("Improperly encoded query string received: '%s'. "
-                         "Proceeding, but did you mean '%s'?" %
-                         (querystr, urllib.parse.urlencode(pairs)))
-                    warnings.warn(s, UserWarning)
-
         items = []
-        parsed_items = urllib.parse.parse_qsl(querystr, keep_blank_values=True)
-        for (key, value), pairstr in six.moves.zip(parsed_items, pairstrs):
-            # Empty value without '=', like '?sup'.
-            if key == quote_plus(utf8(pairstr)):
-                value = None
-            items.append((key, value))
+
+        pairstrs = [s2 for s1 in querystr.split('&') for s2 in s1.split(';')]
+        pairs = [item.split('=', 1) for item in pairstrs]
+        pairs = [(p[0], lget(p, 1, '')) for p in pairs]  # Pad with value ''.
+
+        for pairstr, (key, value) in six.moves.zip(pairstrs, pairs):
+            valid_key = is_valid_encoded_query_key(key)
+            valid_value = is_valid_encoded_query_value(value)
+            if self.strict and (not valid_key or not valid_value):
+                msg = (
+                    "Incorrectly percent encoded query string received: '%s'. "
+                    "Proceeding, but did you mean '%s'?" %
+                    (querystr, urllib.parse.urlencode(pairs)))
+                warnings.warn(msg, UserWarning)
+
+            key_decoded = unquote(key.replace('+', ' '))
+            # Empty value without a '=', e.g. '?sup'.
+            if key == pairstr:
+                value_decoded = None
+            else:
+                value_decoded = unquote(value.replace('+', ' '))
+
+            items.append((key_decoded, value_decoded))
+
         return items
 
 
@@ -1166,7 +1233,7 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         <username>. Initially None.
       scheme: URL scheme. A string ('http', 'https', '', etc) or None.
         All lowercase. Initially None.
-      host: URL host (domain, IPv4 address, or IPv6 address), not
+      host: URL host (hostname, IPv4 address, or IPv6 address), not
         including port. All lowercase. Initially None.
       port: Port. Valid port values are 1-65535, or None meaning no port
         specified.
@@ -1244,12 +1311,12 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
             host is not None and lget(host, 0) == '[' and ':' in host and
             lget(host, -1) == ']')
         if (host is not None and not resembles_ipv6_literal and
-           not is_valid_domain(host)):
+           not is_valid_host(host)):
             errmsg = (
                 "Invalid host '%s'. Host strings must have at least one "
                 "non-period character, can't contain any of '%s', and can't "
                 "have adjacent periods.")
-            raise ValueError(errmsg % (host, INVALID_DOMAIN_CHARS))
+            raise ValueError(errmsg % (host, INVALID_HOST_CHARS))
 
         if callable_attr(host, 'lower'):
             host = host.lower()
@@ -1285,12 +1352,14 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         if userpass or self.username is not None:
             userpass += '@'
 
-        netloc = idna_encode(self.host) or ''
+        netloc = idna_encode(self.host)
         if self.port and self.port != DEFAULT_PORTS.get(self.scheme):
-            netloc += ':' + str(self.port)
+            netloc = (netloc or '') + (':' + str(self.port))
 
-        netloc = (userpass or '') + (netloc or '')
-        return netloc if (netloc or self.host == '') else None
+        if userpass or netloc:
+            netloc = (userpass or '') + (netloc or '')
+
+        return netloc
 
     @netloc.setter
     def netloc(self, netloc):
@@ -1305,14 +1374,14 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
 
         username = password = host = port = None
 
-        if '@' in netloc:
+        if netloc and '@' in netloc:
             userpass, netloc = netloc.split('@', 1)
             if ':' in userpass:
                 username, password = userpass.split(':', 1)
             else:
                 username = userpass
 
-        if ':' in netloc:
+        if netloc and ':' in netloc:
             # IPv6 address literal.
             if ']' in netloc:
                 colonpos, bracketpos = netloc.rfind(':'), netloc.rfind(']')
@@ -1332,7 +1401,7 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         # that if an exception is raised when assigning self.port,
         # self.host isn't updated.
         self.port = port  # Raises ValueError on invalid port.
-        self.host = host or None
+        self.host = host
         self.username = None if username is None else unquote(username)
         self.password = None if password is None else unquote(password)
 
@@ -1365,7 +1434,7 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
 
     @url.setter
     def url(self, url):
-        return self._parse(url)
+        return self.load(url)
 
     def add(self, args=_absent, path=_absent, fragment_path=_absent,
             fragment_args=_absent, query_params=_absent):
@@ -1559,8 +1628,10 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
           path: A list of path segments to remove from the end of the
             existing path segments list, or a path string to remove from
             the end of the existing path string, or True to remove the
-            path entirely.
-          query: If True, remove the query portion of the URL entirely.
+            path portion of the URL entirely.
+          query: A list of query keys to remove from the query, if they
+            exist, or True to remove the query portion of the URL
+            entirely.
           query_params: A list of query keys to remove from the query,
             if they exist.
           port: If True, remove the port from the network location
@@ -1576,26 +1647,29 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
           password: If True, remove the password, if it exists.
         Returns: <self>.
         """
-        if port is True:
-            self.port = None
         if username is True:
             self.username = None
         if password is True:
             self.password = None
+        if port is True:
+            self.port = None
         if path is not _absent:
             self.path.remove(path)
+
         if args is not _absent:
             self.query.remove(args)
         if query is not _absent:
             self.query.remove(query)
-        if fragment is not _absent:
-            self.fragment.remove(fragment)
         if query_params is not _absent:
             self.query.remove(query_params)
+
+        if fragment is not _absent:
+            self.fragment.remove(fragment)
         if fragment_path is not _absent:
             self.fragment.path.remove(fragment_path)
         if fragment_args is not _absent:
             self.fragment.query.remove(fragment_args)
+
         return self
 
     def tostr(self, query_delimiter='&', query_quote_plus=True):
@@ -1607,19 +1681,16 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
             str(self.fragment),
         ))
 
-        # Special cases.
-        if self.scheme is None:
-            if url.startswith('//'):
-                url = url[2:]
-            elif url.startswith('://'):
-                url = url[3:]
-        elif self.scheme in COLON_SEPARATED_SCHEMES:
-            # Change a '://' separator to ':'. Leave a ':' separator as-is.
-            url = _set_scheme(url, self.scheme)
-        elif (self.scheme is not None and
-              (url == '' or  # Protocol relative URL.
-               (url == '%s:' % self.scheme and not str(self.path)))):
-            url += '//'
+        # Differentiate between '' and None values for scheme and netloc.
+        if self.scheme == '':
+            url = ':' + url
+
+        if self.netloc == '':
+            if self.scheme is None:
+                url = '//' + url
+            elif strip_scheme(url) == '':
+                url = url + '//'
+
         return str(url)
 
     def join(self, *urls):
